GameplaySettings{
		usepuzzlegrid = true,
		puzzlerows = 7,
		puzzlecols = 3,
        greypercent=0, -- we manualy override this in OnTrafficCreated
        colorcount=1,
        usetraffic = true,
        automatic_traffic_collisions = false, -- the game shouldn't check for block collisions since we'll be doing that ourselves in this script
        jumpmode="none",
        --matchcollectionseconds=1.5,
        matchcollectionseconds=1,
		forcecollectiononoverfill=true,
		minmatchsize=3,
        greyaction="permanent", -- "eraseall"  -- "eraseblock"

		--track generation settings
		gravity=-.6, -- even without jumping the gravity setting is (a little bit) relevant. It's used in generating the track to sculpt it steep enough to allow jumps
        playerminspeed = 0.1,--so the player is always moving somewhat
        --playermaxspeed = 3
		playermaxspeed = 4,
        minimumbestjumptime = 2.5,--massage the track until a jump of at least this duration is possible
        --uphilltiltscaler = 2,--set to 1 for normal track. higher for steeper
        --downhilltiltscaler = 2,--set to 1 for normal track. higher for steeper
        uphilltiltscaler = 1.6,--set to 1 for normal track. higher for steeper
        downhilltiltscaler = 1.6,--set to 1 for normal track. higher for steeper
        --uphilltiltsmoother = 0.03,
        --uphilltiltsmoother = 0.12,
        uphilltiltsmoother = 0.01, --0.02
        --downhilltiltsmoother = 0.06,
        --downhilltiltsmoother = 0.12,
        downhilltiltsmoother = 0.01,
        useadvancedsteepalgorithm = false,--set false for a less extreme track
        alldownhill = false,
		trafficcompression=0.52,
		--end track generation settings
}

SetSkinProperties{
	lanedividers={-1.5,1.5},
	shoulderlines={-4.5,4.5},
	trackwidth = 5,
	prefersteep = true
}

powernodes = powernodes or {}
powernodetypes = powernodetypes or {}

track = track or {}

function deepcopy(orig) -- a pretty standard lua function for making a copy of a table
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

player={
	score=0,
	prevInput={},
	iPrevRing=0,
	hasFinishedScoringPrevRing=false,
	uniqueName = "Player",
	num=1,
	prevFirstBlockCollisionTested = 1,
	pos = {0,0,0},
	posCosmetic = {0,0,1.5},
	scale = {1.5,1.5,1.5},
	controller = "mouse",
	points = 0 -- used for accumulating points this player earns at each match collection. temp var
}

function hashTraffic(traffic)
	theBigNumber = 0
	for i=1, #traffic do
		theBigNumber = theBigNumber + ((traffic[i].chainend - traffic[i].chainstart + 1) * traffic[i].strength)
	end
	
	return theBigNumber
end

lanespace = 3
half_lanespace = 1.5

blocks = blocks or {}
blockNodes = blockNodes or {}
blockOffsets = blockOffsets or {}

-- since the game makes a track completely diffirent with lots of colors, lets make our own algoritm
function OnTrafficCreated(traffic)
	half_lanespace = lanespace / 2
    
    math.randomseed(hashTraffic(traffic))
    
    local newTraffic = {}
	local avgStrength = 0
    local avgStrengthDecay = 0.9
    
    for i = 1, #traffic do
        local block = traffic[i]
        avgStrength = avgStrengthDecay*avgStrength + (1-avgStrengthDecay)*block.strength
        
		local newBlock = cloneBlock(block)
		
        if (block.strength > avgStrength) then
            newBlock.type = 6
            newBlock.strafe = newBlock.lane*lanespace
            addBlock(newBlock, newTraffic)
        else
            newBlock.type = 5
            newBlock.chainstart = newBlock.impactnode
            newBlock.chainend = newBlock.impactnode
            newBlock.lane = math.random(-1, 1)
            
            addBlock(newBlock, newTraffic)
        end
        
        if (i < #traffic) then
            refBlock = traffic[i+1]
            local refStart;
            if (refBlock.type == 5) then
                refStart = refBlock.impactnode
            else
                refStart = refBlock.chainstart
            end
            local distance = refStart - newBlock.chainend
            local amount = math.floor(distance/5)
            if (amount > 8) then
                amount = 8
            end
            
            for j = 1, amount-1 do
                newBlock = cloneBlock(block)
                newBlock.lane = math.random(-1, 1)
                newBlock.impactnode = math.floor(block.impactnode + distance*j/amount)
                newBlock.chainstart = newBlock.impactnode
                newBlock.chainend = newBlock.impactnode
                addBlock(newBlock, newTraffic)
            end
        end
	end
    
    saveCleanFinish(newTraffic)

    return newTraffic
end

function saveCleanFinish(newTraffic)
    newTraffic[#newTraffic-2].type = 6
    newTraffic[#newTraffic-1].type = 6
    newTraffic[#newTraffic-0].type = 6
    local i = math.random(0,2)
    newTraffic[#newTraffic-i].lane = 0
    newTraffic[#newTraffic-i].strafe = 0*lanespace
end

function addBlock(newBlock, newTraffic)
    newBlock.strafe = newBlock.lane*lanespace
    newTraffic[#newTraffic+1] = newBlock
    blocks[#blocks+1]=newBlock
    blockNodes[#blockNodes+1] = newBlock.impactnode
    blockOffsets[#blockOffsets+1] = {newBlock.lane*lanespace, 0, 0}
end

function cloneBlock(block)
    local newBlock = {}
    
    newBlock.type = 5
    newBlock.impactnode = block.impactnode
    newBlock.chainstart = block.chainstart
    newBlock.chainend = block.chainend
    newBlock.lane = block.lane
    newBlock.strength = block.strength
    newBlock.strafe = block.strafe
    
    return newBlock
end

beakerPos = {0,0,1.5}
beakerScale = {.7,.7,.7}
function OnSkinLoaded()-- called after OnTrafficCreated. The skin script has loaded content.
	CreateClone{name="beaker", prefabName="Vehicle", attachToTrackWithNodeOffset=-1, transform={pos=beakerPos,scale=beakerScale}}
	SetScoreboardNote{text="STEALTH"}
end


alreadyremoving = false
function RemoveAllGreys()
	if not alreadyremoving then
		alreadyremoving = true --just some ensurane
		local countremoved = 0
		local puzzle = GetPuzzle()
		local cells = puzzle["cells"]
		for colnum=1,#cells do
			local col = cells[colnum]
			for rownum=1,#col do
				local cell = col[rownum]
				if cell.type == 5 then
					cells[colnum][rownum].type = -1
					countremoved = countremoved + 1
				end
			end
		end
		SetPuzzle{cells=cells}

		--score penalty for shuriken
		if countremoved > 0 then
			score = score - 777
			if score < 0 then
				score = 0
			end
			SetGlobalScore{score=score,showdelta=true}
		end
		alreadyremoving = false
	end
end

function PuzzleCollecting_Groupvalue(count)
	if count == 3 then return 120
	elseif count == 4 then return 243
	elseif count == 5 then return 398
	elseif count == 6 then return 580
	elseif count == 7 then return 786
	elseif count == 8 then return 1010
	elseif count == 9 then return 1251
	elseif count == 10 then return 1504
	elseif count == 11 then return 1768
	elseif count == 12 then return 2041
	elseif count == 13 then return 2316
	elseif count == 14 then return 2585
	elseif count == 15 then return 2846
	elseif count == 16 then return 3097
	elseif count == 17 then return 3337
	elseif count == 18 then return 3567
	elseif count == 19 then return 3787
	elseif count == 20 then return 3997
	elseif count == 21 then return 4200
	else return 0 end
end


score = 0 --the global score (in multiplayer, shared by all players co-operatively)
combo = 0 --consecutive colors hit

function OnPuzzleCollecting()
	local points = 0
	local currentmatch = 0

	local puzzle = GetPuzzle()
	local matchSize = puzzle["matchedcellscount"]
	if matchSize < 1 then -- no matches collected. Drop chain bonus here

	else
		local cells = puzzle["cells"]
		for colnum=1,#cells do
			local col = cells[colnum]
			for rownum=1,#col do
				local cell = col[rownum]
				if cell["matched"] then
					local cellPoints = 10 * ((cell["type"]+1) * cell["matchsize"])
					currentmatch = currentmatch + 1
				end
			end
		end
		points = PuzzleCollecting_Groupvalue(currentmatch)
		score = score + points -- add the shared points earned from this collection batch
								
		SetGlobalScore{score=score,showdelta=true}
	end
end

iCurrentRing = 0 --Update function keeps this current
blocksToHide = {}
stealthy = true

function Collide(strafe, hitAllLanes)
	local playerLane = 0;
	if strafe>half_lanespace then playerLane = 1
	elseif strafe<-half_lanespace then playerLane=-1 end

	local maxRing = iCurrentRing + 2
	local foundFirst = false
	for i=player.prevFirstBlockCollisionTested,#blockNodes do
		if not blocks[i].tested then
			if blockNodes[i] <= maxRing then
				if not foundFirst then
					player.prevFirstBlockCollisionTested = i
					foundFirst = true
				end

				if blockNodes[i] <= math.floor(iCurrentRing) then
					if not blocks[i].hidden then
                        if (blocks[i].lane == playerLane) or hitAllLanes then
                            blocksToHide[#blocksToHide+1] = i
                            blocks[i].hidden = true
                            local blockOffset = blockOffsets[i]
                            if blocks[i].type == 5 then
                                if stealthy then
                                    stealthy = false
                                    SetScoreboardNote{text="", color={200,200,200}}
                                end
                                if stealthy or combo > 0 then
                                    PlaySound{name="landcrash"} -- play crash sound when the combo resets
                                end
                                combo = 0
                                
                                SetPuzzle{newblocks={{type=5, collision_strafe=blockOffset[1], puzzle_col=blocks[i].lane+1, add_top=true}}}
                            else
                                combo = combo + 1
                                points = (combo * 4) - 3
                                if points > 200 then
                                    points = 200
                                end
                                score = score + points
                                local pitch = 1 + 2 * GetPercentPuzzleFilled(false)
                                PlaySound{name="hit",pitch=pitch}
                                --SetScoreboardNote{text=stealth_text, color={200,200,200}, secondsvisible=0}
                                SetGlobalScore{score=score,showdelta=true}
                                SetPuzzle{newblocks={{type=blocks[i].type, collision_strafe=blockOffset[1], puzzle_col=blocks[i].lane+1, add_top=true}}}
                                --SetPuzzle{newblocks={{type=0, collision_strafe=blockOffset[1], puzzle_col=blocks[i].lane+1, add_top=true}}}
                            end
                            --SetPuzzle{newblocks={{type=blocks[i].type, collision_strafe=blockOffset[1], puzzle_col=blocks[i].lane+1, add_top=true}}}
                            SendCommand{command="HoverUp"} -- causes the hovering vehicle to bounce up a bit
                        end
					end
					blocks[i].tested = true
				end
			else
				break --stop the loop once we get to a block way past the player
			end
			--end
		end
	end
	
	--if #blocksToHide > 0 then UpdateBatchRenderer{uniqueName="BlockBatchRenderer", hideLocations=blocksToHide} end
end


function GetPercentPuzzleFilled(countMatchedCellsAsEmpty)
	local puzzle = GetPuzzle()
	local cells = puzzle["cells"]
	local numFilled = 0
	local numCells = 0;
	for colnum=1,#cells do
		local col = cells[colnum]
		for rownum=1,#col do
			numCells = numCells + 1
			local cell = col[rownum]
			if cell.type >= 0 then
				if not (countMatchedCellsAsEmpty and cell["matched"]) then
					numFilled = numFilled + 1
				end
			end
		end
	end

	return numFilled / numCells
end

function GetNextCollisionStrafe() -- figure out which lane the minion should be in
	for i=player.prevFirstBlockCollisionTested,#blockNodes do
		if not blocks[i].tested then
			return blocks[i].lane * lanespace
		end
	end

	return 0
end

prevHitAllLanes = true
playerPos = {0,0,1.5}
doHitAllLanes = false
function Update(dt, tracklocation, playerstrafe, input) --called every frame
	--iCurrentRing = math.floor(GetCurrentTrackLocation())
	--local input = GetInput()
	iCurrentRing = math.floor(tracklocation)
	local playersInput = input["players"]
	local mouseInput = input["mouse"]
	
	if dt>0 then
		doHitAllLanes = false
		if playersInput["Button 1"] or mouseInput["LMB"] then
			doHitAllLanes = true
		end

		if playersInput["Button 2"] or mouseInput["RMB"] then
			RemoveAllGreys()
		end
	end
    
	if prevHitAllLanes ~= doHitAllLanes then
		if doHitAllLanes then
			SendCommand{command="Show", name="beaker"}
		else
			SendCommand{command="Hide", name="beaker"}
		end

		prevHitAllLanes = doHitAllLanes
	end
	
	player.posCosmetic[1] = playerstrafe
	player.pos[1] = playerstrafe
	
	--move the helper to line it up for the next block hit
	local desiredHelperStrafe = GetNextCollisionStrafe()
	beakerPos[1] = beakerPos[1] + dt * 11 * (desiredHelperStrafe - beakerPos[1])
	SendCommand{command="SetTransform", name="beaker", param={pos=beakerPos,scale=beakerScale}}

	blocksToHide = {}
	
	Collide(playerstrafe, doHitAllLanes)
	
	if #blocksToHide > 0 then
		HideTraffic(blocksToHide)
	end
end

function OnRequestFinalScoring()
	local cleanFinishBonus = 0
	local percentPuzzleFilledAndUnmatched = GetPercentPuzzleFilled(true)
	if percentPuzzleFilledAndUnmatched == 0 then
		cleanFinishBonus = math.floor(score * .25)
	end
	
	local stealthBonus = 0
	if stealthy then 
		stealthBonus = math.floor(score * .3) 
	end

	return {
		rawscore = score,
		bonuses = {
			"Stealth Bonus (30%): "..stealthBonus,
			"Clean Finish (25%): "..cleanFinishBonus,
		},
		finalscore = score + stealthBonus + cleanFinishBonus
	}
end